{-
MDULO INTERCODEMONAD:
  Mdulo com a definio e as operaes sobre o Monad InterCode

Equipe:
  Andr Wilson Brotto Furtado
  Gustavo Danzi de Andrade
  Monique Louise de Barros Monteiro
-}

-- definio e funes mondicas para o Monad InterCode
module InterCodeMonad where

import Lexparser
import Util

-- Os tipos de ns da rvore de cdigo intermedirio
-- elementos nao terminais...
data InterElement
    = AtribBin InterTerminal InterOp InterTerminal InterTerminal
    | AtribUni InterTerminal InterOp InterTerminal
    | Copia InterTerminal InterTerminal
    | Call InterTerminal [InterTerminal]
    | Label String
    | GoFalse String
    | GoTrue String
    | Goto String
    | Box InterTerminal
    | Pop InterTerminal
    | Push InterTerminal

-- elementos terminais...
data InterOp
       = OpMais
       | OpMenos
       | OpEqual
       | OpAnd
       | OpOr
       | OpNot
       | OpConcat
       | OpLength

data InterTerminal
       = TermInt Int
       | TermBool Bool
       | TermString String
       | TermTemp Int
       | TermArg Int
       | TermFunction String Type -- nome original da funo / tipo de retorno

-- Implementao do show dos tipos de dados do cdigo intermedirio
instance Show InterElement where
  show (AtribBin t1 op t2 t3) = "    " ++ show(t1) ++ " = " ++ show(t2) ++ " " ++ show(op) ++ " " ++ show(t3)
  show (AtribUni t1 op t2) = "    " ++ show(t1) ++ " = " ++ " " ++ show(op) ++ " " ++ show(t2)
  show (Copia t1 t2) = "    " ++ show(t1) ++ " = " ++ show(t2)
  show (Call term params) = "    " ++ "call " ++ show(term) ++ "(" ++ (show params) ++ ")"
  show (Label s) = s ++ ":"
  show (GoFalse s) = "    " ++ "GoFalse " ++ " " ++ s
  show (GoTrue s) = "    " ++ "GoTrue " ++ s
  show (Goto s) = "    " ++ "Goto " ++ s
  show (Box t) = "    " ++ "box " ++ show(t)
  show (Pop t) = "    " ++ "Pop " ++ show(t)
  show (Push t                    ) = "    " ++ "Push " ++ show(t)  

instance  Show InterOp where
  show (OpMais    ) = "+"
  show (OpMenos   ) = "-"
  show (OpEqual   ) = "="
  show (OpAnd       ) = "AND"
  show (OpOr        ) = "OR"
  show (OpNot       ) = "NOT"
  show (OpConcat    ) = "++"
  show (OpLength      ) = "LENGTH"

instance  Show InterTerminal where
  show (TermInt  i)   = show(i)
  show (TermBool   b)   = show(b)
  show (TermString s)   = s
  show (TermTemp   n)   = "rhcil" ++ show(n)
  show (TermArg    n)   = "arg" ++ show(n)
  show (TermFunction orignalName fType) = orignalName  ++ "::" ++ (show fType)

data FunctionInfo = FunctionInfo {
  name                    :: String,
  params                  :: [String],
  codeList                :: [InterElement],
  returnType              :: Type,
  nextTemp                :: Int,
  nextArgs                :: Int,
  dataTypes               :: [Type],
  renameList              :: [(String, InterTerminal)]     -- associa um ident a um TermTemp
}      

data InterCodeState = InterCodeState {
  nextLabel            :: Int,
  renameFunctionsList  :: [(String, InterTerminal)],      -- associa um identidicador de funo a um TermFunction
  functions            :: [FunctionInfo],                 -- funes da aplicao
  functionContext      :: [FunctionInfo]           -- pilha cujo topo  a funo na qual o cdigo intermedirio deve ser gerado
}

-- Definies padres de Monads
data InterCode a = IC ( InterCodeState -> (InterCodeState, a) )

instance Monad InterCode
  where
    -- (>>=) :: InterCodeState a -> (a -> InterCodeState b) -> InterCodeState b
    (IC p) >>= k  =  IC( \s0 -> let (s1, a) = p s0
                                    (IC q) = k a
                                in q s1 )

    -- return :: a -> InterCodeState a
    return a = IC( \interCode -> (interCode, a) )

---------------------------------------------------------------------------------
--                             FUNES MONDICAS                               --
---------------------------------------------------------------------------------
-- funcao de aplicacao do Monad
applyInterCode :: InterCode () -> InterCodeState -> (InterCodeState,())
applyInterCode (IC f) interCodeState = f interCodeState

getNextTemp :: InterCode Int
getNextTemp = IC(\interCode ->
   let newListTemp = getNextTempAux (functionContext interCode) in
   (interCode {functionContext = fst newListTemp}, snd newListTemp ))

getNextTempAux :: [FunctionInfo] -> ([FunctionInfo],Int)
getNextTempAux [] = error "InterCodeMonad.getNextTempAux error: FunctionInfo list is not empty"
getNextTempAux (f:[]) =
   let newTemp = (nextTemp f) + 1 in
   ([f{nextTemp = newTemp}],newTemp - 1) -- -1 porque ele no deve retornar o valor incrementado
getNextTempAux (f:fs) = (f : (fst newListTemp), snd newListTemp)
   where newListTemp = getNextTempAux fs
   
getNextArgs :: InterCode Int
getNextArgs = IC(\interCode ->
   let newListArgs = getNextArgsAux (functionContext interCode) in
   (interCode {functionContext = fst newListArgs}, snd newListArgs))

getNextArgsAux :: [FunctionInfo] -> ([FunctionInfo],Int)
getNextArgsAux [] = error "InterCodeMonad.getNextArgsAux error: FunctionInfo list is not empty"
getNextArgsAux (f:[]) =
   let newArgs = (nextArgs f) + 1 in
   ([f{nextArgs = newArgs}],newArgs - 1) -- -1 porque ele no deve retornar o valor incrementado
getNextArgsAux (f:fs) = (f : (fst newListArgs), snd newListArgs)
   where newListArgs = getNextArgsAux fs

getNextLabel :: InterCode String
getNextLabel = IC(\interCode ->
 (interCode {nextLabel = (nextLabel interCode) + 1}, "Label"++(show $ nextLabel interCode)))

addToDataType :: Type -> InterCode ()
addToDataType typeElement = IC(\interCode -> (interCode {functionContext = addToDataTypeAux typeElement (functionContext interCode)},()))

addToDataTypeAux :: Type -> [FunctionInfo] -> [FunctionInfo]
addToDataTypeAux _ [] = []
addToDataTypeAux t (f:[]) = [f{dataTypes = (dataTypes f) ++ [t]}]
addToDataTypeAux t (f:fs) = f:(addToDataTypeAux t fs)

addFunctionInfo :: FunctionInfo -> InterCode ()
addFunctionInfo fInfo = IC(\interCode -> (interCode {functions = (functions interCode) ++ [fInfo]},()))

-- Recupera as informaes de uma funo a partir do seu nome (retorna a ltima encontrada)
getFunctionInfo :: String -> InterCode FunctionInfo
getFunctionInfo funcName = IC(\interCode -> (interCode, getFunctionInfoAux funcName ((reverse (functions interCode)) ++ (reverse (functionContext interCode)))))

getFunctionInfoAux :: String -> [FunctionInfo] -> FunctionInfo
getFunctionInfoAux funcName [] = error ("InterCodeMonad.getFunctionInfoAux error: Nao foi encontrado nenhuma funcao de nome " ++ funcName)
getFunctionInfoAux funcName (fInfo:fis)
                   | funcName == (name fInfo) = fInfo
                   | otherwise = getFunctionInfoAux funcName fis

addFunctionContext :: FunctionInfo -> InterCode ()
addFunctionContext fInfo = IC(\interCode -> (interCode {functionContext = (functionContext interCode) ++ [fInfo]},()))

removeFunctionContext :: InterCode ()
removeFunctionContext = IC(\interCode -> let
    lastFInfo = getLastContext interCode
    rest = take (length (functionContext interCode) -1) (functionContext interCode) in
    (interCode {functions = (functions interCode) ++ [lastFInfo], functionContext = rest},()))

-- retorna o topo da pilha de contextos
getLastContext :: InterCodeState -> FunctionInfo
getLastContext interCode = last (functionContext interCode)

addToInterCode :: InterElement -> InterCode ()
addToInterCode interElement = IC(\interCode -> (interCode {functionContext = addToInterCodeAux interElement (functionContext interCode)},()))

addToInterCodeAux :: InterElement -> [FunctionInfo] -> [FunctionInfo]
addToInterCodeAux _ [] = error "InterCodeMonad.addToInterCodeAux error: addToInterCode but function list is empty"
addToInterCodeAux interElement (f:[]) = [f{codeList = (codeList f) ++ [interElement]}]
addToInterCodeAux interElement (f:fs) = f:(addToInterCodeAux interElement fs)

addToRenameList :: (String, InterTerminal) -> InterCode ()
addToRenameList renamedElement = IC(\interCode ->
   (interCode {functionContext = addToRenameListAux renamedElement (functionContext interCode)},()))

addMultipleToRenameList :: [String] -> InterCode ()
addMultipleToRenameList [] = return ()
addMultipleToRenameList (x:xs) = do
   n <- getNextTemp
   addToRenameList (x, (TermTemp n))
   addMultipleToRenameList xs

-- Adiciona os parametros na lista de renomeacao
addParametersToRenameList :: [String] -> InterCode ()
addParametersToRenameList [] = return ()
addParametersToRenameList (x:xs) = do
   n <- getNextArgs
   addToRenameList (x, (TermArg n))
   addParametersToRenameList xs

addToRenameListAux :: (String, InterTerminal) -> [FunctionInfo] -> [FunctionInfo]
addToRenameListAux _ [] = []
addToRenameListAux renamedElement (f:[]) = [f{renameList = (renameList f) ++ [renamedElement]}]
addToRenameListAux renamedElement (f:fs) = f:(addToRenameListAux renamedElement fs)

getFromRenameList :: String -> InterCode InterTerminal
getFromRenameList ident = IC(\interCode ->
    (interCode, getFromRenameListAuxList ident (reverse (functionContext interCode))))

getFromRenameListAuxList :: String -> [FunctionInfo] -> InterTerminal
getFromRenameListAuxList ident [] = error ("InterCodeMonad.getFromRenameListAuxList error: Nao foi encontrado nenhum ident de nome " ++ ident)
getFromRenameListAuxList ident (a:as)
        | (isInContext ident (renameList a)) = getFromRenameListAux ident (renameList a)
        | otherwise = getFromRenameListAuxList ident as

isInContext :: String -> [(String, InterTerminal)] -> Bool
isInContext ident [] = False
isInContext ident ((identName, identTerm):is)
       | (ident == identName) = True
       | otherwise = isInContext ident is

getFromRenameListAux :: String -> [(String, InterTerminal)] -> InterTerminal
getFromRenameListAux ident [] = error ("InterCodeMonad.getFromRenameListAux error: Nao foi encontrado nenhum ident de nome " ++ ident)
getFromRenameListAux ident ((identName, identTerm):is)
       | (ident == identName) = identTerm
       | otherwise = getFromRenameListAux ident is

addToRenameFunctionsList :: (String, InterTerminal) -> InterCode ()
addToRenameFunctionsList renamedFunction = IC(\interCode -> (interCode {renameFunctionsList = (renameFunctionsList interCode) ++ [renamedFunction]},()))

getFromRenameFunctionsList :: String -> InterCode InterTerminal
getFromRenameFunctionsList ident = IC(\interCode -> (interCode, getFromRenameFunctionsListAux ident (reverse (renameFunctionsList interCode))))

getFromRenameFunctionsListAux :: String -> [(String, InterTerminal)] -> InterTerminal
getFromRenameFunctionsListAux ident [] = error ("getFromRenameFunctionsAux.error: Nao foi encontrada nenhuma funcao com o nome " ++ ident)
getFromRenameFunctionsListAux ident ((identName, identTerm):is)
                              | (ident == identName) = identTerm
                              | otherwise = getFromRenameListAux ident is

-- Funes auxiliares
getOpType :: InterOp -> Type
getOpType OpMais = TypeInt
getOpType OpMenos = TypeInt
getOpType OpEqual = TypeBool
getOpType OpAnd = TypeBool
getOpType OpOr = TypeBool
getOpType OpNot = TypeBool
getOpType OpConcat = TypeString
getOpType OpLength = TypeString

----------------------------------------------------------------------------------
--      INFERNCIA DO TIPO DO PROGRAMA PARA A GERAO DE CDIGO FINAL            --
----------------------------------------------------------------------------------
getProgramType :: Programa -> InterCode Type
getProgramType (Programa exp) = getTypeExpressao exp

getTypeExpressao :: Expressao -> InterCode Type
getTypeExpressao (ExpressaoValor valor) =  getTypeValor valor
getTypeExpressao (ExpressaoUnaria expUnaria) = getTypeExpUnaria expUnaria
getTypeExpressao (ExpressaoBinaria expBinaria) = getTypeExpBinaria expBinaria
getTypeExpressao (ExpressaoDeclaracao expDeclaracao) = getTypeExpDeclaracao expDeclaracao
getTypeExpressao (ExpressaoId ident) = return $ TypeObject
getTypeExpressao (ExpressaoIf expIf) = getTypeExpIf expIf
getTypeExpressao (ExpressaoAplicacao expAplicacao) = getTypeExpAplicacao expAplicacao

getTypeValor :: Valor -> InterCode Type
getTypeValor (Valor (ValorInteiro _)) = return $ TypeInt
getTypeValor (Valor (ValorBooleano _)) = return $ TypeBool
getTypeValor (Valor (ValorString _)) = return $ TypeString

getTypeExpUnaria :: ExpUnaria -> InterCode Type
getTypeExpUnaria (PreMenos _) = return $ TypeInt
getTypeExpUnaria (Not _) = return $ TypeBool
getTypeExpUnaria (Length _) = return $ TypeInt

getTypeExpBinaria :: ExpBinaria -> InterCode Type
getTypeExpBinaria (Mais _ _) = return $ TypeInt
getTypeExpBinaria (Menos _ _) = return $ TypeInt
getTypeExpBinaria (And _ _) = return $ TypeBool
getTypeExpBinaria (Or _ _) = return $ TypeBool
getTypeExpBinaria (Igual _ _) = return $ TypeBool
getTypeExpBinaria (Concatenacao _ _) = return $ TypeString

getTypeExpDeclaracao :: ExpDeclaracao -> InterCode Type
getTypeExpDeclaracao (LetIn _ exp) = getTypeExpressao exp

getTypeExpIf :: ExpIf -> InterCode Type
getTypeExpIf (IfThenElse _ expThen expElse) = do
	typeThen <- getTypeExpressao expThen
	typeElse <- getTypeExpressao expElse
	return $ chooseType typeThen typeElse

chooseType :: Type -> Type -> Type
chooseType TypeUnknown TypeUnknown = error "InterCodeMonad.chooseType error: TypeUnknown"
chooseType TypeUnknown t = t
chooseType t TypeUnknown = t
chooseType t1 t2 
	| (t1 == t2) = t1
	| otherwise = error "InterCodeMonad.chooseType error: as expresses de Then e Else de um If devem ser do mesmo tipo!"

getTypeExpAplicacao :: ExpAplicacao -> InterCode Type
getTypeExpAplicacao (Aplicacao fName _) = getFunctionType fName

getFunctionType :: String -> InterCode Type
getFunctionType fName = IC(\interCode -> (interCode, getFunctionTypeAux fName (reverse (functions interCode))))



getFunctionTypeAux :: String -> [FunctionInfo] -> Type
getFunctionTypeAux fName [] = TypeUnknown -- error ("InterCodeMonad.getFunctionTypeAux error: Nao foi encontrada nenhuma funcao com o nome " ++ fName)
getFunctionTypeAux fName (fInfo:fs)
                              | (name fInfo == fName) = (returnType fInfo)
                              | otherwise = getFunctionTypeAux fName fs

                              
changeMainFunctionType :: Type -> InterCode ()
changeMainFunctionType newType = IC(\interCode ->
	let mainFunction = head (functionContext interCode)
	    otherFunctions = tail (functionContext interCode) in
	(interCode {functionContext = (mainFunction{returnType = newType}):otherFunctions},()))

changeFunctionType :: Type -> InterCode ()
changeFunctionType newType = IC(\interCode -> (interCode {functionContext = changeFunctionTypeAux newType (functionContext interCode)},()))

changeFunctionTypeAux :: Type -> [FunctionInfo] -> [FunctionInfo]
changeFunctionTypeAux _ [] = error "InterCodeMonad.changeFunctionTypeAux error: changeFunctionTypeAux but function list is empty"
changeFunctionTypeAux newType (f:[]) = [f{returnType = newType}]
changeFunctionTypeAux newType (f:fs) = f:(changeFunctionTypeAux newType fs)
	
-- FOR TESTING PURPOSES                       

listFunctions :: InterCode String
listFunctions = IC(\interCode -> (interCode, listFunctionsAux (functions interCode) [] ))

listFunctionsAux :: [FunctionInfo] -> String -> String
listFunctionsAux [] accum = error accum
listFunctionsAux (f:fs) accum = listFunctionsAux fs (accum ++ ", " ++ (name f))
